# Code adapted from from django.utils.http.py
import re
import datetime
import calendar


# list of months
MONTHS = "jan feb mar apr may jun jul aug sep oct nov dec".split()

# two digit day: 22, 10, 01, 02, etc
__D = r"(?P<day>\d{2})"

# also two digit day: 25, 01, etc
__D2 = r"(?P<day>[ \d]\d)"

# month: jan, fev, mar...
__M = r"(?P<mon>\w{3})"

# year, 4 digits: 1995, 2021
__Y = r"(?P<year>\d{4})"

# year, 2 digits: 95, 87
__Y2 = r"(?P<year>\d{2})"

# Timestamp
__T = r"(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})"


# first three characters are the name of the day: Mon, Tue, Wed, Thu, Fri, Sat,
# Sun ex: Sun, 06 Nov 1994 08:49:37 GMT
RFC1123_DATE = re.compile(r"^\w{3}, %s %s %s %s GMT$" % (__D, __M, __Y, __T))

# here the day name can be Monday, Tuesday, Wednesday, Thursday, Saturday, 
# Sunday hence \w{6,9} in the beginning
# ex: Sunday, 06-Nov-94 08:49:37 GMT
RFC850_DATE = re.compile(r"^\w{6,9}, %s-%s-%s %s GMT$" % (__D, __M, __Y2, __T))

# ex: Sun Nov  6 08:49:37 1994
ASCTIME_DATE = re.compile(r"^\w{3} %s %s %s %s$" % (__M, __D2, __T, __Y))


def parse_http_date(date: str) -> int:
    # TODO: parse_http_date is originaly from django.utils.http.py
    """
    Note: this code is originally from django.utils.http.py
    Parse a date format as specified by HTTP RFC7231 section 7.1.1.1.

    The three formats allowed by the RFC are accepted, even if only the first
    one is still in widespread use.

    Return an integer expressed in seconds since the epoch, in UTC.
    """
    # email.utils.parsedate() does the job for RFC1123 dates; unfortunately
    # RFC7231 makes it mandatory to support RFC850 dates too. So we roll
    # our own RFC-compliant parsing.
    for regex in RFC1123_DATE, RFC850_DATE, ASCTIME_DATE:
        m = regex.match(date)
        if m is not None:
            break
    else:
        raise ValueError("%r is not in a valid HTTP date format" % date)
    try:
        year = int(m["year"])
        if year < 100:
            current_year = datetime.datetime.now(datetime.timezone.utc).year
            current_century = current_year - (current_year % 100)
            if year - (current_year % 100) > 50:
                # year that appears to be more than 50 years in the future are
                # interpreted as representing the past.
                year += current_century - 100
            else:
                year += current_century
        _month = MONTHS.index(m["mon"].lower()) + 1
        _day = int(m["day"])
        _hour = int(m["hour"])
        _min = int(m["min"])
        _sec = int(m["sec"])
        result = datetime.datetime(year, _month, _day, _hour, _min, _sec)
        return calendar.timegm(result.utctimetuple())
    except Exception as exc:
        raise ValueError("%r is not a valid date" % date) from exc



from base64 import b64decode, b64encode
from easy_hmac import exceptions
from easy_hmac.utils import http
from datetime import datetime, timezone, timedelta
import hashlib
import hmac


def generate_hmac_sha256(
    secret: str, method: str, body: str, path: str, timestamp: str
):
    """Generate a SHA256 HMAC from two strings

    Args:
        secret: the key to use for the HMAC
        method: the http method. Likely POST.
        body: the data to generate the digest from
        path: the remote path. /remote/webhook/path
        timestamp: datetimestamp, GMT formatted as "%a, %d %b %Y %H:%M:%S GMT"
    Returns:
        str: the HMAC digest.
    """

    contentType: str = "application/json"
    _hash = hashlib.md5(body.encode())
    contentMD5 = b64encode(_hash.digest()).decode("utf-8")
    message_parts = [method, contentMD5, contentType, timestamp, path]
    message = "\n".join(message_parts)
    signature = hmac.new(
        bytes(secret, "latin-1"), bytes(message, "latin-1"), digestmod=hashlib.sha256
    )
    return signature.digest()


def __verify_hmac(secret: str, hmac_bytes: bytes, message_string: str) -> bool:
    """
    Verifies if the hmac generated by message_string and secret is the same as hmac_bytes.
    This function is used by verify_hmac

    Args:
        secret: the key to use for the HMAC
        hmac_bytes: the hmac to be verified
        message_string: the message to use for the HMAC

    Returns:
        boolean: True if the HMACs are equal, False otherwise
    """

    message_bytes: bytes = message_string.encode("utf-8")
    secret_bytes: bytes = secret.encode("utf-8")
    correct_hmac_bytes: bytes = hmac.new(
        secret_bytes, msg=message_bytes, digestmod=hashlib.sha256
    ).digest()
    return hmac.compare_digest(hmac_bytes, correct_hmac_bytes)


def verify_hmac(
    secret: str,
    hmac_base64: str,
    md5_body: str,
    raw_body: bytes,
    timestamp: str,
    content_type: str,
    path: str,
    request_method: str,
):
    """
    Verifies if the hmac generated by the parameters and secret is the same as
    the one by hmac_base64.

    Args:
        secret: the key to use for the HMAC
        hmac_base64: a given hmac
        md5_body:
        raw_body:
        timestamp:
        content_type:
        path:
        request_method:

    Returns:
        boolean: True if the HMACs are equal, False otherwise
    """

    # check if timestamp is valid before verifying the HMAC
    try:
        request_time = datetime.fromtimestamp(
            http.parse_http_date(timestamp), timezone.utc
        )
    except ValueError as e:
        raise exceptions.AuthenticationFailed("Malformed Date header") from e

    if abs(request_time - datetime.now(timezone.utc)) > timedelta(minutes=15):
        raise exceptions.AuthenticationFailed("Request time too old")

    # define auth_failed here so we can raise the exception easily
    auth_failed = exceptions.AuthenticationFailed("HMAC Authentication Failed")

    # ensure that the base64 md5 hash computed from the ENCODED raw body is the 
    # same as the one present on md5_body
    body_hash_b64 = b64encode(hashlib.new("md5", raw_body).digest())\
        .decode("utf-8")

    if body_hash_b64 != md5_body:
        raise auth_failed

    hmac_bytes = b64decode(hmac_base64)

    # builds the message
    message_parts = [
        request_method,
        body_hash_b64,
        content_type,
        timestamp,
        path,
    ]

    message = "\n".join(message_parts)

    if __verify_hmac(secret, hmac_bytes, message):
        return True

    raise auth_failed



# Test suite for easy_hmac
from base64 import b64encode
from datetime import datetime
from easy_hmac import core, exceptions
import hashlib
import hmac
import unittest


class TestEasyHmac(unittest.TestCase):
    """
    Houses the tests for the easy_hmac package
    """

    def compute_dummy_request(self, token, body, a_timestamp, application_id):
        """
        Computes a dummy request for testing purposes
        """

        method = "GET"
        path = f"/api/v1/applications/{application_id}/bundle"
        timestamp = a_timestamp
        contentType = "application/json"

        hash = hashlib.md5(body.encode())
        contentMD5 = b64encode(hash.digest()).decode("utf-8")
        message_parts = [method, contentMD5, contentType, timestamp, path]
        message = "\n".join(message_parts)

        signature = hmac.new(
            bytes(token["secret"], "latin-1"),
            bytes(message, "latin-1"),
            digestmod=hashlib.sha256,
        )
        hmac_base64 = b64encode(signature.digest()).decode("utf-8")

        headers = {
            "Date": timestamp,
            "Content-MD5": contentMD5,
            "Content-Type": contentType,
            "Authorization": "HMAC {}:{}".format(token["identifier"], hmac_base64),
        }

        return {"method": method, "body": body, "path": path, "headers": headers}

    def test_verify_hmac(self):
        """
        Tests if verify_hmac actually verifies the HMAC generated by a dummy request
        """

        token = {
            "name": "adriano",
            "secret": "6ff2dc141c0841e2a43c25be9ae9b097",
            "identifier": "2e42a19593f047e080285e49864b0fb6",
        }

        timestamp = datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        request = TestEasyHmac.compute_dummy_request(token, "", timestamp, 1234)
        body = request["body"]
        headers = request["headers"]
        path = request["path"]
        method = request["method"]

        # preparing test_request parameters
        auth_header = headers["Authorization"]
        auth_method, auth_string = auth_header.split(" ")
        token_identifier, hmac_base64 = auth_string.split(":")

        actual = core.verify_hmac(
            token["secret"],
            hmac_base64,
            headers["Content-MD5"],
            body.encode(),
            headers["Date"],
            headers["Content-Type"],
            path,
            method,
        )

        self.assertTrue(actual)

    def test_verify_hmac_wrong_secret(self):
        """
        Tests if verify_hmac raises exception if secret is wrong
        """

        token = {
            "name": "adriano",
            "secret": "6ff2dc141c0841e2a43c25be9ae9b097",
            "identifier": "2e42a19593f047e080285e49864b0fb6",
        }

        timestamp = datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        request = TestEasyHmac.compute_dummy_request(token, "", timestamp, 1234)
        body = request["body"]
        headers = request["headers"]
        path = request["path"]
        method = request["method"]

        # preparing test_request parameters
        auth_header = headers["Authorization"]
        auth_method, auth_string = auth_header.split(" ")
        token_identifier, hmac_base64 = auth_string.split(":")

        wrong_secret = "Adriano"
        self.assertRaises(
            exceptions.AuthenticationFailed,
            core.verify_hmac,
            wrong_secret,
            hmac_base64,
            headers["Content-MD5"],
            body.encode(),
            headers["Date"],
            headers["Content-Type"],
            path,
            method,
        )

    def test_verify_hmac_malformed_timestamp(self):
        """
        Tests if verify_hmac raises exception if timestamp is malformed
        """

        token = {
            "name": "adriano",
            "secret": "6ff2dc141c0841e2a43c25be9ae9b097",
            "identifier": "2e42a19593f047e080285e49864b0fb6",
        }

        malformed_timestamp = "Sunday Nov 6 08:49:37 1994"
        request = TestEasyHmac.compute_dummy_request(
            token, "", malformed_timestamp, 1234
        )
        body = request["body"]
        headers = request["headers"]
        path = request["path"]
        method = request["method"]

        # preparing test_request parameters
        auth_header = headers["Authorization"]
        auth_method, auth_string = auth_header.split(" ")
        token_identifier, hmac_base64 = auth_string.split(":")

        # malformed_date = "Sun Nov  6 08:49:37 1994"
        self.assertRaises(
            exceptions.AuthenticationFailed,
            core.verify_hmac,
            token["secret"],
            hmac_base64,
            headers["Content-MD5"],
            body.encode(),
            headers["Date"],
            headers["Content-Type"],
            path,
            method,
        )

    def test_verify_hmac_timestamp_too_old(self):
        """
        Tests if verify_hmac raises exception if timestamp is too old
        """

        token = {
            "name": "adriano",
            "secret": "6ff2dc141c0841e2a43c25be9ae9b097",
            "identifier": "2e42a19593f047e080285e49864b0fb6",
        }

        old_timestamp = "Fri, 10 Dec 2021 00:16:57 GMT"
        request = TestEasyHmac.compute_dummy_request(token, "", old_timestamp, 1234)
        body = request["body"]
        headers = request["headers"]
        path = request["path"]
        method = request["method"]

        # preparing test_request parameters
        auth_header = headers["Authorization"]
        auth_method, auth_string = auth_header.split(" ")
        token_identifier, hmac_base64 = auth_string.split(":")

        self.assertRaises(
            exceptions.AuthenticationFailed,
            core.verify_hmac,
            token["secret"],
            hmac_base64,
            headers["Content-MD5"],
            body.encode(),
            headers["Date"],
            headers["Content-Type"],
            path,
            method,
        )

    def test_verify_hmac_wrong_body(self):
        """
        Tests if verify_hmac raises exception if the body of the message is different is wrong
        """

        token = {
            "name": "adriano",
            "secret": "6ff2dc141c0841e2a43c25be9ae9b097",
            "identifier": "2e42a19593f047e080285e49864b0fb6",
        }

        timestamp = datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        request = TestEasyHmac.compute_dummy_request(token, "", timestamp, 1234)
        headers = request["headers"]
        path = request["path"]
        method = request["method"]

        # preparing test_request parameters
        auth_header = headers["Authorization"]
        auth_method, auth_string = auth_header.split(" ")
        token_identifier, hmac_base64 = auth_string.split(":")

        wrong_body = "{'hello': 'world'}"
        self.assertRaises(
            exceptions.AuthenticationFailed,
            core.verify_hmac,
            token["secret"],
            hmac_base64,
            headers["Content-MD5"],
            wrong_body.encode(),
            headers["Date"],
            headers["Content-Type"],
            path,
            method,
        )

    def test_generate_hmac(self):
        """
        Tests if generate_hmac_sha256 will generate the same HMAC generated by a sample webhook
        given the parameters are the same
        """

        secret = "79721503-d1ef-46b7-b4ca-fec39ece902f"
        body = '{"event": "crawler.lifecycle_updated", "payload": {"uuid": "cb8c79cd-8d79-4698-90a2-662eeab8da98", "vendor_specific_id": "1639095349767-4727fc80", "vendor_label": "adriano", "bank_id": 130, "timestamp": "2021-12-10T00:16:08.048401Z", "status": "PROCESSING", "sequence": -1}}'
        secret = secret
        method = "POST"
        timestamp = "Fri, 10 Dec 2021 00:16:57 GMT"
        path = ""
        result_digest = core.generate_hmac_sha256(secret, method, body, path, timestamp)
        actual_signature = b64encode(result_digest).decode()
        expected_signature = "SLtkg8653hvTwT0onBWI47vky0O65kkL99XN/cUskXc="
        self.assertEqual(
            expected_signature,
            actual_signature,
            f"Expected = {expected_signature} but got Actual = {actual_signature}",
        )


if __name__ == "__main__":
    unittest.main()



# Test suite for easy_hmac
from easy_hmac.utils.http import parse_http_date
import unittest


class TestUtilsHTTP(unittest.TestCase):
    """
    Houses the tests for the easy_hmac.utils.http
    """

    def test_parse_http_date_wrong_date_format(self):
        """
        Tests if a date on a format not accepted by parse_http_date
        raises a ValueError exception
        """
        wrong_date = "Monday, 14/12/2021 - 10:47:23"
        self.assertRaises(ValueError, parse_http_date, wrong_date)

    def test_parse_http_date_RFC1123(self):
        """
        Given a date on the RFC1123 format, tests if it correctly parses to epoch
        Epoch value, 784111777, was given by https://www.epochconverter.com/
        """
        RFC1123 = "Sun, 06 Nov 1994 08:49:37 GMT"
        actual = parse_http_date(RFC1123)
        self.assertEqual(784111777, actual)

    def test_parse_http_date_RFC850(self):
        """
        Given a date on the RFC850 format, tests if it correctly parses to epoch
        Epoch value, 784111777, was given by https://www.epochconverter.com/
        """
        RFC850_date = "Sunday, 06-Nov-94 08:49:37 GMT"
        actual = parse_http_date(RFC850_date)
        self.assertEqual(784111777, actual)

    def test_parse_http_date_ASCII(self):
        """
        Given a date on the ASCII format, tests if it correctly parses to epoch
        Epoch value, 784111777, was given by https://www.epochconverter.com/
        """
        ASCII_date = "Sun Nov  6 08:49:37 1994"
        actual = parse_http_date(ASCII_date)
        self.assertEqual(784111777, actual)


if __name__ == "__main__":
    unittest.main()
